local hideBlizz = CreateFrame("Frame")
hideBlizz:Hide()
hideBlizz:SetAlpha(0)

local _hori = {
	"LEFT",
	"CENTER",
	"RIGHT",
}

local _vert = {
	"TOP",
	"MIDDLE",
	"BOTTOM",
}

local anchors = {
	"TopLeft",
	"Top",
	"TopRight",
	"Right",
	"BottomRight",
	"Bottom",
	"BottomLeft",
	"Left",
	"Center",
}

local layers = {
	'BACKGROUND',
	'LOW',
	'MEDIUM',
	'HIGH',
	'DIALOG',
	'FULLSCREEN',
	'FULLSCREEN_DIALOG',
	'TOOLTIP',
}

local components = {
	{
		name = "box",
		create = function(frame)
			local name = frame:GetName().."Box"
			local f = _G[name] or CreateFrame("Button", name, frame, "SecureUnitButtonTemplate")
			f:EnableMouse(false)
			f:SetPoint("Center")
			f.unit = frame.unit
			f:SetAttribute("unit", frame.unit)
			RegisterUnitWatch(f)
			f.unit = frame.unit
			return f
		end,
		functions = {
			Load = function(self)
				--self:Layout()
			end,
			Layout = function(self)
				self:SetSize(self:GetParent().sets.width, self:GetParent().sets.height)
			end,
			
		},
		scripts = {
			OnEnter = function(self)
				self.unit = self:GetParent().unit
				UnitFrame_OnEnter(self)
			end,
			OnLeave = function(self)
				UnitFrame_OnLeave(self)
			end,
			OnUpdate = function(self, elapsed)
				local parts = self:GetParent().components
				if not parts then
					return
				end
				for i , part in pairs(parts) do
					if part.Update then
						part:Update(elapsed)
					end
				end
			end,
		},
	},--box
	{
		name = "drop",
		create = function(frame)
			local name = frame:GetName().."DropDown"
			local f = _G[name] or CreateFrame("Frame",name, frame.components.box, "UIDropDownMenuTemplate")
			f.unit = frame.unit
			frame.components.box.dropDown = f
			frame.components.box:SetAttribute("*type1", "target");
			frame.components.box:SetAttribute("*type2", "menu");
			frame.components.box:RegisterForClicks("LeftButtonUp", "RightButtonUp");
			return f
		end,
		functions = {
			Load = function(self)
				if self.Initialize then
					CompactUnitFrame_SetMenuFunc(self:GetParent(), self.Initialize);
				end
			end,
			Initialize = function(self)
				local menu;
				local name;
				local unit = self.unit
				local id = nil;
				if ( UnitIsUnit(unit, "player") ) then
					menu = "SELF";
				elseif ( UnitIsUnit(unit, "vehicle") ) then
					menu = "VEHICLE";
				elseif ( UnitIsUnit(unit, "pet") ) then
					menu = "PET";
				elseif ( UnitIsOtherPlayersBattlePet(unit) ) then
					menu = "OTHERBATTLEPET";
				elseif ( UnitIsBattlePet(unit) ) then
					menu = "BATTLEPET";
				elseif ( UnitIsOtherPlayersPet(unit) ) then
					menu = "OTHERPET";	
				elseif ( UnitIsPlayer(unit) ) then
					id = UnitInRaid(unit);
					if ( id ) then
						menu = "RAID_PLAYER";
					elseif ( UnitInParty(unit) ) then
						menu = "PARTY";
					else
						menu = "PLAYER";
					end
				else
					menu = "TARGET";
					name = RAID_TARGET_ICON;
				end
				if ( menu ) then
					return UnitPopup_ShowMenu(self, "_"..menu, unit, name, id);
				end
			end,
		},
	},--drop
	{
		name = "back",
		create = function(frame)
			local name = frame:GetName().."Backdrop"
			local f = _G[name] or CreateFrame("Button", name, frame.components.box)
			f:SetAllPoints(frame.components.box)
			f:SetFrameLevel(1)
			f:Show()
			return f
		end,
		functions = {
			Load = function(self)

			end,
			Layout = function(self)
				self:SetBackdrop({
				bgFile = self.sets.bgFile,
					insets = {left = -self.sets.inset, right = -self.sets.inset, top = -self.sets.inset, bottom = -self.sets.inset},
					tile = false,
				})
				self:SetBackdropColor(self.sets.color.r, self.sets.color.g, self.sets.color.b, self.sets.color.a)
				if self.sets.enable == true then
						self:SetAlpha(1)
				else
					self:SetAlpha(0)
				end
			end,
		},
		options = {
			{
				name = "Enable",
				kind = "CheckButton",
				key = "enable",
			},
			{
				name = "Color",
				kind = "ColorPicker",
				key = "color",
			},
			{
				name = "Padding",
				kind = "Slider",
				key = "inset",
				min = -30,
				max = 30,
				step = 1,
			},
		},
		defaults = {
			bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
			bgcolor = {
				a = 100,
				r = 100,
				g = 100,
				b = 100,
			},
			color = {
				a = 1,
				b = 0,
				g = 0.411764705882353,
				r = 0.0509803921568627,
			},
			bgPad = 0,
			edgecolor = {
				a = 100,
				r = 100,
				g = 100,
				b = 100,
			},
			edgeFile = "Interface\\ACHIEVEMENTFRAME\\UI-Achievement-WoodBorder",
			enable = true,
			inset = 3,
			edgeSize = 13,
		},
	},--back
	{
		name = "name",
		create = function(frame)
			local name = frame:GetName().."Name_control"
			local control = _G[name] or CreateFrame("Frame", name, frame.components.box)
			control:SetFrameLevel(frame:GetFrameLevel()+20)
			local name = frame:GetName().."Name"
			local f = _G[name] or control:CreateFontString(name, 'ARTWORK', 'GameFontNormal')
			f:SetAllPoints(control)
			control.text = f
			return control
		end,
		functions = {
			Load = function(self)

			end,
			Layout = function(self)
				local sets = self.sets
				if sets.enable ~= true then
					self:Hide()
					self.noUpdate = true
					self.text:SetText("")
					return
				else
					self:Show()
					self.noUpdate = nil
				end
				self:Reposition()
				self:SetSize(sets.width, sets.height)
				self.text:SetJustifyH(_hori[sets.justifyH]) -- Sets horizontal text justification ("LEFT","RIGHT", or "CENTER")
				self.text:SetJustifyV(_vert[sets.justifyV]) -- Sets vertical text justification ("TOP","BOTTOM", or "MIDDLE")
				self:Rescale()
			end,
			Update = function(self)
				if self.noUpdate then
					return
				end
				local unit = self:GetParent().unit
				self.text:SetText(UnitName(unit))
			end,
			Rescale = function(self)
				self:SetScale(self.sets.scale/100)
			end,
			Reposition = function(self)
				self:ClearAllPoints()
				local x = self.sets.x/(self.sets.scale/100)
				local y = self.sets.y/(self.sets.scale/100)
				self:SetPoint(self.sets.anchor, self:GetParent(), x, y)
			end,
		},
		options = {
			{
				name = "Enable",
				kind = "CheckButton",
				key = "enable",
			},
			{
				name = "Anchor",
				kind = "Menu",
				key = "anchor",
				table = anchors,
			},
			{
				name = "Justify Vert",
				kind = "Slider",
				key = "justifyV",
				min = 1,
				max = 3,
			},
			{
				name = "Justify Hori",
				kind = "Slider",
				key = "justifyH",
				min = 1,
				max = 3,
			},
			{
				name = "X Offset",
				kind = "Slider",
				key = "x",
				min = -400,
				max = 400,
			},
			{
				name = "Y Offset",
				kind = "Slider",
				key = "y",
				min = -400,
				max = 400,
			},
			{
				name = "Width",
				kind = "Slider",
				key = "width",
				min = 10,
				max = 400,
				step = 1,
			},
			{
				name = "Height",
				kind = "Slider",
				key = "height",
				min = 12,
				max = 240,
				step = 14,
			},
			{
				name = "Scale",
				kind = "Slider",
				key = "scale",
				min = 25,
				max = 250,
			},
		},
		defaults = {
			enable = true,
			justifyV = 1,
			width = 85,
			y = -17,
			x = -20,
			justifyH = 1,
			anchor = "TopRight",
			height = 26,
			scale = 100,
		},
	},--name
	{
		name = "level",
		create = function(frame)
			local name = frame:GetName().."Level_control"
			local control = _G[name] or CreateFrame("Frame", name, frame.components.box)
			control:SetFrameLevel(frame:GetFrameLevel()+10)
			local name = frame:GetName().."Level"
			local f = _G[name] or control:CreateFontString(name, 'ARTWORK', 'GameFontNormal')
			f:SetAllPoints(control)
			control.text = f
			control.unit = frame.unit
			return control
		end,
		functions = {
			Load = function(self)

			end,
			Layout = function(self)
				local sets = self.sets
				local text = self.text
				if sets.enable ~= true then
					self:Hide()
					self.noUpdate = true
					text:SetText("")
					return
				else
					self:Show()
					self.noUpdate = nil
				end
				self:ClearAllPoints()
				self:SetPoint(sets.anchor, self:GetParent(), sets.x, sets.y)
				self:SetSize(sets.width, sets.height)
				self:SetScale(sets.scale/100)
		

				self.text:SetJustifyH(_hori[sets.justifyH]) -- Sets horizontal text justification ("LEFT","RIGHT", or "CENTER")
				self.text:SetJustifyV(_vert[sets.justifyV]) -- Sets vertical   text justification ("TOP","BOTTOM", or "MIDDLE")
			end,
			Update = function(self)
				if self.noUpdate then
					return
				end
				local oldLevel = self.oldLevel
				local level = UnitLevel(self.unit)
				if level == -1 then
					level = "??"
				elseif level == 0 then
					level = ""
				end
				if level ~= oldLevel then --doesn't call "SetText" unless it has too. saves memory
					self.oldLevel = level
					self.text:SetText(level)
				end
			end,
		},
		options = {
			{
				name = "Anchor",
				kind = "Menu",
				key = "anchor",
				table = anchors,
			},
			{
				name = "Enable",
				kind = "CheckButton",
				key = "enable",
			},
			{
				name = "Justify Vert",
				kind = "Slider",
				key = "justifyV",
				min = 1,
				max = 3,
			},
			{
				name = "Justify Hori",
				kind = "Slider",
				key = "justifyH",
				min = 1,
				max = 3,
			},
			{
				name = "X Offset",
				kind = "Slider",
				key = "x",
				min = -400,
				max = 400,
			},
			{
				name = "Y Offset",
				kind = "Slider",
				key = "y",
				min = -400,
				max = 400,
			},
			{
				name = "Width",
				kind = "Slider",
				key = "width",
				min = 17,
				max = 400,
				step = 1,
			},
			{
				name = "Height",
				kind = "Slider",
				key = "height",
				min = 12,
				max = 400,
				step = 11,
			},
			{
				name = "Scale",
				kind = "Slider",
				key = "scale",
				min = 50,
				max = 250,
			},
		},
		defaults = {
			scale = 100,
			justifyV = 1,
			width = 25,
			y = -17,
			x = 0,
			justifyH = 3,
			height = 12,
			enable = true,
			anchor = "TopRight",
		},
	},--level
	{
		name = "cast",
		create = function(frame)
			local name = frame:GetName().."Cast_control"
			local opac = _G[name] or CreateFrame("Frame", name, frame.components.box)--opacity control
			local name = frame:GetName().."Cast"
			local f = _G[name] or CreateFrame("StatusBar",  name, opac, "CastingBarFrameTemplate")
			f.opac = opac
			f:Hide()
			f:SetAttribute("unit", frame.unit)
			CastingBarFrame_SetLook(f, "CLASSIC")
			f.text:SetHeight(12)
			f.unit = frame.unit
			f.time = f.time or f:CreateFontString(nil, "OVERLAY", "TextStatusBarText")
			f.time:SetTextColor(1.0,1.0,1.0)
			f.time:SetAllPoints(f)
			f.drop = f.drop or CreateFrame("Frame", f:GetName().."backdrop", frame.components.box)
			f.drop:SetPoint("TopLeft", f)
			f.drop:SetPoint("BottomRight", f)
			
			f.border:SetParent(hideBlizz)
			f.borderShield:SetParent(hideBlizz)
			f.icon:SetParent(hideBlizz)
			f.barFlash:SetParent(hideBlizz)
			return f
		end,
		functions = {
			Load = function(self)
				CastingBarFrame_OnLoad (self, self.unit)
				self:Show()
			end,
			Layout = function(self)
				local sets = self.sets
				if sets.enable ~= true then
					self:Hide()
					self.noUpdate = true
					return
				else
					--self:Show()
					self.noUpdate = nil
				end
				self:ClearAllPoints()
				self:SetPoint(sets.anchor, self:GetParent():GetParent(), sets.x, sets.y)
				self:SetSize(sets.width, sets.height)
				if sets.scale < 25 then
					sets.scale = sets.scale *100
				end
				self:SetScale(sets.scale/100)
				self.opac:SetAlpha(sets.alpha/100)
				local anchor = sets.anchor
				local hori = "CENTER" -- "LEFT","RIGHT", or "CENTER"
				local vert = "MIDDLE" -- "TOP","BOTTOM", or "MIDDLE"
				local H, V = "RIGHT", "MIDDLE"
		
				self.text:ClearAllPoints()
				self.text:SetPoint(anchor)
		
		
				if string.find(anchor, "Left") then
					hori = "LEFT"
					H = "RIGHT"
				elseif string.find(anchor, "Right") then
					hori = "RIGHT"
					H = "LEFT"
				end
		
				if string.find(anchor, "Top") then
					vert = "TOP"
					V = "TOP"
				elseif string.find(anchor, "Bottom") then
					vert = "BOTTOM"
					V = "BOTTOM"
				end

				if V == "MIDDLE" then
					V = "CENTER"
				end

			self.text:SetJustifyH(hori) -- Sets horizontal text justification ("LEFT","RIGHT", or "CENTER")
			self.text:SetWidth(self:GetWidth()-24)
			self.time:SetJustifyH(H)
			self.time:SetJustifyV(V)
			self.text:SetHeight(sets.textheight)
			if self.sets.showBG then
				self.drop:SetBackdrop(self.bg)
			else
				self.drop:SetBackdrop(nil)
			end
			local lay = layers[sets.frameLayer]
			self:SetFrameStrata(lay)
			self:SetFrameLevel(sets.frameLevel)
			self.drop:SetFrameStrata(lay)
			self.drop:SetFrameLevel(sets.frameLevel)
		end,
			Update = function(self, elapsed)
				if self.noUpdate then
					return
				end
				local curGUID = UnitGUID(self.unit)
				if (curGUID ~= self.lastGUID) or not curGUID then
					if self.lastGUID then
						self:GetScript("OnShow")(self)
					end
					if not curGUID then
						return
					end
				end
				
				local kind, name, subText, text, texture, startTime, endTime, _,  castID = self:isCasting()
				self.seconds = self.seconds or ""
								
				if kind then
					--Events to mimmick
						--cast start
						--cast delayed
						--channel Start
					if kind == "cast" then
						if not self.startTime then 
							CastingBarFrame_OnEvent (self, "UNIT_SPELLCAST_START", self.unit, name, nil, castID, select( 7, GetSpellInfo(name)))
						else
							if self.endTime ~= endTime then
								CastingBarFrame_OnEvent (self, "UNIT_SPELLCAST_DELAYED", self.unit, self.spellname, nil, self.castID, self.spellID)
							end
						end
					elseif kind == "channel" then
						CastingBarFrame_OnEvent (self, "UNIT_SPELLCAST_CHANNEL_START", self.unit, name, 0, select( 7, GetSpellInfo(name)))
					end

					self.begin = self.begin or GetTime()
					self.kind = kind
					self.spellID = select( 7, GetSpellInfo(name))
					self.startTime = startTime
					self.endTime = endTime
					self.spellname = name
					if kind == "cast" then
						self.castID = castID
					else
						self.castID = castID or 0
					end
					self.stop = nil
				elseif self.kind then
					local time = GetTime()
					if ((time) < ((self.endTime/1000) - .5)) then
						if self.kind == "cast" then
							self.casting = true
							CastingBarFrame_OnEvent (self, "UNIT_SPELLCAST_INTERRUPTED", self.unit, self.spellname, nil, self.castID, self.spellID)
						elseif self.kind == "channel" then
							self.channeling = true
							CastingBarFrame_OnEvent (self, "UNIT_SPELLCAST_CHANNEL_STOP", self.unit, self.spellname, nil, self.castID, self.spellID)
						end
					else
						if self.kind == "cast" then
							self.casting = true
							CastingBarFrame_OnEvent (self, "UNIT_SPELLCAST_STOP", self.unit, self.spellname, nil, self.castID, self.spellID)
						elseif self.kind == "channel" then
							self.channeling = true
							CastingBarFrame_OnEvent (self, "UNIT_SPELLCAST_CHANNEL_STOP", self.unit, self.spellname, nil, self.castID, self.spellID)
						end
					end

					self.kind = nil
					self.spellID = nil
					self.startTime = nil
					self.endTime = nil
					self.spellname = nil
					self.begin = nil
					self.seconds = ""
					self.stop = true
				end
				if self.stop then
					self.time:SetText("")
				else
					if self.endTime then
						self.seconds = string.format("%.1f" ,(self.endTime / 1000) - GetTime())
						self.time:SetText(self.seconds)
					end
				end
				self.lastGUID = curGUID
			end,
			isCasting = function(self)
				if UnitCastingInfo(self.unit) then
					return "cast", UnitCastingInfo(self.unit)
				elseif UnitChannelInfo(self.unit) then
					return "channel", UnitChannelInfo(self.unit)
				end
			end,
			bg = {
				bgFile = 'Interface\\Tooltips\\UI-Tooltip-Background',
				insets = {left = 0, right = 0, top = 0, bottom = 0},
				tile = false,
			},
		},
		scripts = {
			OnEvent = function(self)
			
			end,
			OnShow = function(self)
				local kind,_, _, _, _, startTime, endTime = self:isCasting()
	
				if (kind == "cast") then
					if ( startTime ) then
						self.value = (GetTime() - (startTime / 1000));
					end
				elseif (kind == "channel") then
					if ( endTime ) then
						self.value = ((endTime / 1000) - GetTime());
					end
				else
					self.fadeOut = 0
					self.holdTime = 0
					self:Hide()
				end
			end,
		-- [[

			OnEnter = function(self)
				local name = UnitCastingInfo(self.unit) or UnitChannelInfo(self.unit)
				if name then
					local spellId = select(7, GetSpellInfo(name))
					if spellId then
						GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
						GameTooltip:ClearAllPoints()
						GameTooltip:SetPoint("BottomRight", self, "TopLeft")
						GameTooltip:SetSpellByID(spellId)
					end
				end
			end,
			OnLeave = function(self)
				GameTooltip:Hide()
			end,
		--]]
		},
		options = {
			{
				name = "Anchor",
				kind = "Menu",
				key = "anchor",
				table = anchors,
			},
			{
				name = "Frame Level",
				kind = "Slider",
				key = "frameLevel",
				min = 1,
				max = 100,
			},	
			{
				name = "Frame Layer",
				kind = "Slider",
				key = "frameLayer",
				min = 1,
				max = 8,
			},
			{
				name = "Enable",
				kind = "CheckButton",
				key = "enable",
			},
			{
				name = "Text Height",
				kind = "Slider",
				key = "textheight",
				min = 12,
				max = 240,
				step = 14,
			},
			{
				name = "X Offset",
				kind = "Slider",
				key = "x",
				min = -400,
				max = 400,
			},
			{
				name = "Y Offset",
				kind = "Slider",
				key = "y",
				min = -400,
				max = 400,
			},
			{
				name = "Width",
				kind = "Slider",
				key = "width",
				min = 10,
				max = 400,
			},
			{
				name = "Height",
				kind = "Slider",
				key = "height",
				min = 10,
				max = 400,
			},
			{
				name = "Scale",
				kind = "Slider",
				key = "scale",
				min = 25,
				max = 250,
			},
			{
				name = "Opacity",
				kind = "Slider",
				key = "alpha",
				min = 0,
				max = 100,
			},	
			{
				name = "Backdrop",
				kind = "CheckButton",
				key = "showBG",
			},
		},
		defaults = {
			frameLevel = 2,
			frameLayer = 3,
			scale = 100,
			alpha = 100,
			width = 106,
			y = 18,
			x = 0,
			anchor = "BottomRight",
			height = 34,
			enable = true,
			showBG = false,
			textheight = 26,
		},
	},--cast
	{
		name = "health",
		create = function(frame)
			local name = frame:GetName().."Health"
			local f = _G[name] or CreateFrame("StatusBar",  name, frame.components.box, "TextStatusBar")
			f:SetStatusBarTexture("Interface\\RaidFrame\\Raid-Bar-Hp-Fill", "BORDER")
			f:SetStatusBarColor(0,1,0,1)
			f:EnableMouse(false)
			f:Show()
			f.text = f.text or f:CreateFontString(nil, "OVERLAY", "GameFontNormal")
			f.text:SetTextColor(1.0,1.0,1.0)
			f.text:SetAllPoints(true)
			return f
		end,
		functions = {
			Load = function(self)
			end,
			Layout = function(self)
				local sets = self.sets
				if sets.enable ~= true then
					self:Hide()
					self.noUpdate = true
					return
				else
					self:Show()
					self.noUpdate = nil
				end
				self:Reposition()
				self:Rescale()
				self:SetSize(sets.width, sets.height)
				self.text:SetJustifyH(sets.justifyH) -- Sets horizontal text justification ("LEFT","RIGHT", or "CENTER")
				self.text:SetJustifyV(sets.justifyV) -- Sets vertical   text justification ("TOP","BOTTOM", or "MIDDLE")
				self:SetAlpha(sets.alpha/100)
				if sets.drop then
					self:SetBackdrop( {bgFile = "Interface\\QUESTFRAME\\UI-TextBackground-BackdropBackground"})
				else
					self:SetBackdrop(nil)
				end
			end,
			Update = function(self)
				if self.noUpdate then
					return
				end
				local unit = self:GetParent().unit
				local max = UnitHealthMax(unit)
				local current = UnitHealth(unit)
				local dead = UnitIsGhost(unit) or UnitIsDead(unit)
				local sets = self.sets
				if dead then
					self:SetMinMaxValues(0, 1)
					self:SetValue(1)
					self.text:SetText("Dead")
				else
					self:SetMinMaxValues(0, max)
					self:SetValue(current)
		
					local value = floor(current)
					if sets.percent then
						self.text:SetText(string.format("%.0f" ,(value/max) * 100) .. "%")
					else
						self.text:SetText(value)
					end
				end
				local r, g, b;
				if not UnitIsConnected(unit) then
					r, g, b = .5, .5, .5
				else
					local localizedClass, englishClass = UnitClass(unit);
					local classColor = RAID_CLASS_COLORS[englishClass];
					if ( classColor ) then --and self:GetParent().sets.useClassColors ) then
						r, g, b = classColor.r, classColor.g, classColor.b;
					else
						if ( UnitIsFriend("player", unit) ) then
							r, g, b = 0.0, 1.0, 0.0;
						else
							r, g, b = 1.0, 0.0, 0.0;
						end
					end
				end
				if dead then
					r, g, b = r-.75, g-.75, b-.75
				end
				if ( r ~= self.r or g ~= self.g or b ~= self.b) then
					self:SetStatusBarColor(r, g, b);
					self.r, self.g, self.b = r, g, b;
				end
			end,
			Rescale = function(self)
				self:SetScale(self.sets.scale/100)
			end,
			Reposition = function(self)
				self:ClearAllPoints()
				local x = self.sets.x/(self.sets.scale/100)
				local y = self.sets.y/(self.sets.scale/100)
				self:SetPoint(self.sets.anchor, self:GetParent(), x, y)
			end,
		},
		options = {
			{
				name = "Anchor",
				kind = "Menu",
				key = "anchor",
				table = anchors,
			},
			{
				name = "Enable",
				kind = "CheckButton",
				key = "enable",
			},
			{
				name = "Percent",
				kind = "CheckButton",
				key = "percent",
			},
			{
				name = "Backdrop",
				kind = "CheckButton",
				key = "drop",
			},
			{
				name = "Opacity",
				kind = "Slider",
				key = "alpha",
				min = 0,
				max = 100,
			},
			{
				name = "X Offset",
				kind = "Slider",
				key = "x",
				min = -400,
				max = 400,
			},
			{
				name = "Y Offset",
				kind = "Slider",
				key = "y",
				min = -400,
				max = 400,
			},
			{
				name = "Width",
				kind = "Slider",
				key = "width",
				min = 10,
				max = 400,
			},
			{
				name = "Height",
				kind = "Slider",
				key = "height",
				min = 10,
				max = 400,
			},
			{
				name = "Scale",
				kind = "Slider",
				key = "scale",
				min = 25,
				max = 250,
			},		
		},
		defaults = {
			scale = 100,
			height = 18,
			justifyV = "CENTER",
			width = 140,
			y = 0,
			x = 0,
			justifyH = "CENTER",
			drop = true,
			anchor = "TopRight",
			enable = true,
			percent = true,
			alpha = 100,
		},
	},--health
	{
		name = "power",
		create = function(frame)
			local name = frame:GetName().."Power"
			local f = _G[name] or CreateFrame("StatusBar",  name, frame.components.box, "TextStatusBar")
			f:SetStatusBarTexture("Interface\\RaidFrame\\Raid-Bar-Hp-Fill", "BORDER")
			f:SetStatusBarColor(0,1,0,1)
			f:SetBackdrop( {bgFile = "Interface\\QUESTFRAME\\UI-TextBackground-BackdropBackground"})
			f:EnableMouse(false)
			f:Show()
			f.text = f.text or f:CreateFontString(nil, "OVERLAY", "GameFontNormal")
			f.text:SetTextColor(1.0,1.0,1.0)
			f.text:SetAllPoints(true)
			return f
		end,
		functions = {
			Load = function(self)
			end,
			Layout = function(self)
				local sets = self.sets
				if sets.enable ~= true then
					self:Hide()
					self.noUpdate = true
					return
				else
					self:Show()
					self.noUpdate = nil
				end
				local sets = self.sets
				self:Reposition()
				self:Rescale()
				self:SetSize(sets.width, sets.height)
				self.text:SetJustifyH(sets.justifyH) -- Sets horizontal text justification ("LEFT","RIGHT", or "CENTER")
				self.text:SetJustifyV(sets.justifyV) -- Sets vertical   text justification ("TOP","BOTTOM", or "MIDDLE")
				self:SetAlpha(sets.alpha/100)
				if sets.drop then
					self:SetBackdrop( {bgFile = "Interface\\QUESTFRAME\\UI-TextBackground-BackdropBackground"})
				else
					self:SetBackdrop(nil)
				end
			end,
			Update = function(self)
				if self.noUpdate then
					return
				end
				local unit = self:GetParent().unit
				local max = UnitPowerMax(unit)
				local current = UnitPower(unit)
				local dead = UnitIsGhost(unit) or UnitIsDead(unit)
				local sets = self.sets

				if max == 0 or dead then
					self:SetMinMaxValues(0, 1)
					self:SetValue(0)
					self.text:SetText("")
					return
				end 

				if dead then
					self:SetMinMaxValues(0, 1)
					self:SetValue(1)
					self.text:SetText("")
				else
					self:SetMinMaxValues(0, max)
					self:SetValue(current)
					if sets.percent then
						self.text:SetText(string.format("%.0f" ,(current/max) * 100) .. "%")
						else
						self.text:SetText(current)
					end
				end
				local r, g, b
				if ( not UnitIsConnected(unit)) then
					r, g, b = 0.5, 0.5, 0.5
				else
					local powerType, powerToken, altR, altG, altB = UnitPowerType(unit)
					local info = PowerBarColor[powerToken]
					if ( info ) then
						r, g, b = info.r, info.g, info.b
					else
						if ( not altR) then
							info = PowerBarColor[powerType] or PowerBarColor["MANA"]
							r, g, b = info.r, info.g, info.b
						else
							r, g, b = altR, altG, altB
						end
					end
				end
				if dead then
					r, g, b = r-.75, g-.75, b-.75
				end
				self:SetStatusBarColor(r, g, b)
			end,
			Rescale = function(self)
				self:SetScale(self.sets.scale/100)
			end,
			Reposition = function(self)
				self:ClearAllPoints()
				local x = self.sets.x/(self.sets.scale/100)
				local y = self.sets.y/(self.sets.scale/100)
				self:SetPoint(self.sets.anchor, self:GetParent(), x, y)
			end,
		},
		options = {
			{
				name = "Anchor",
				kind = "Menu",
				key = "anchor",
				table = anchors,
			},
			{
				name = "Enable",
				kind = "CheckButton",
				key = "enable",
			},
			{
				name = "Percent",
				kind = "CheckButton",
				key = "percent",
			},
			{
				name = "Backdrop",
				kind = "CheckButton",
				key = "drop",
			},
			{
				name = "Opacity",
				kind = "Slider",
				key = "alpha",
				min = 0,
				max = 100,
			},
			{
				name = "X Offset",
				kind = "Slider",
				key = "x",
				min = -400,
				max = 400,
			},
			{
				name = "Y Offset",
				kind = "Slider",
				key = "y",
				min = -400,
				max = 400,
			},
			{
				name = "Width",
				kind = "Slider",
				key = "width",
				min = 10,
				max = 400,
			},
			{
				name = "Height",
				kind = "Slider",
				key = "height",
				min = 10,
				max = 400,
			},
			{
				name = "Scale",
				kind = "Slider",
				key = "scale",
				min = 25,
				max = 250,
			},		
		},
		defaults = {
			scale = 100,
			height = 18,
			justifyV = "CENTER",
			width = 140,
			y = 0,
			x = 0,
			justifyH = "CENTER",
			drop = true,
			anchor = "Bottom",
			enable = true,
			percent = true,
			alpha = 100,
		},
	},--power
	{
		name = "focus",
		create = function(frame)
			local name = frame:GetName().."Focus"
			local f = _G[name] or CreateFrame("Button", name, frame.components.box, "ActionButtonTemplate, SecureActionButtonTemplate")
			f:SetFrameLevel(frame:GetFrameLevel()+ 10)
			f:SetAttribute("unit", frame.unit)
			f.icon:SetTexture([[Interface\Addons\Dominos_Units\Focus]])
			f:RegisterForClicks("AnyUp")
			f:SetAttribute("*type1", "target");
			
			if frame.unit ~= "focus" then
				f:SetAttribute("*type2", "focus");
				f.Tooltip = "Right-Click to Set Focus."
			else
				f:SetAttribute("*type2", "macro");
				f:SetAttribute("macrotext", '/clearfocus')
				f.Tooltip = "Right-Click to Clear Focus."
			end
			if not f.masqued then
				f.masqued = true
				Dominos:Masque('Focus Macros', f)
			end
			return f
		end,
		functions = {
			Load = function(self)
				self:Show()
			end,
			Layout = function(self)
				local sets = self.sets
				if sets.enable ~= true then
					self:Hide()
					self.noUpdate = true
					self:SetText("")
					return
				else
					self:Show()
					self.noUpdate = nil
				end
				self:ClearAllPoints()
				self:SetPoint(sets.anchor, self:GetParent(), sets.x, sets.y)

				if sets.scale < 25 then
					sets.scale = sets.scale *100
				end
				self:SetScale(sets.scale/100)
				self:SetAlpha(sets.alpha/100)
			end,
			Update = function(self)
				local sets = self.sets
				
				if sets.usePortrait then
					SetPortraitTexture(self.icon, self:GetParent().unit)
 					self.icon:SetVertexColor(1, 1, 1)
				else
					local localizedClass, englishClass = UnitClass(self:GetParent().unit);
					local classColor = RAID_CLASS_COLORS[englishClass];
					if classColor then
						r, g, b = classColor.r, classColor.g, classColor.b;
					else
						r, g, b = 1, 1, 1
					end
					if self.icon:GetTexture() ~= [[Interface\Addons\Dominos_Units\Focus]] then
						self.icon:SetTexture([[Interface\Addons\Dominos_Units\Focus]])
					end
					self.icon:SetVertexColor(r, g, b)
				end
			end,
		},
		scripts = {
			OnEnter = function(self)
				UnitFrame_OnEnter(self:GetParent())
				GameTooltip:AddLine(self.Tooltip)
				GameTooltip:AppendText("")
			end,
			OnLeave = function(self)
				GameTooltip:Hide()
			end,
		},
		options = {
			{
				name = "Anchor",
				kind = "Menu",
				key = "anchor",
				table = anchors,
			},
			{
				name = "Enable",
				kind = "CheckButton",
				key = "enable",
			},
			{
				name = "Opacity",
				kind = "Slider",
				key = "alpha",
				min = 0,
				max = 100,
			},
			{
				name = "X Offset",
				kind = "Slider",
				key = "x",
				min = -400,
				max = 400,
			},
			{
				name = "Y Offset",
				kind = "Slider",
				key = "y",
				min = -400,
				max = 400,
			},
			{
				name = "Scale ",
				kind = "Slider",
				key = "scale",
				min = 25,
				max = 250,
			},
			{
				name = "Portrait Icon",
				kind = "CheckButton",
				key = "usePortrait",
			}
		},
		defaults = {
			y = 0,
			x = 0,
			enable = true,
			anchor = "Left",
			scale = 95,
			alpha = 100,
			usePortrait = true,
		},
	},--focus
	{
		name = "model",
		create = function(frame)
			local name = frame:GetName().."Model"
			local f =  _G[name] or CreateFrame("PlayerModel",  name, frame.components.box, "ModelWithZoomTemplate")
			f:SetFrameLevel(1)
			f:RegisterEvent("UNIT_MODEL_CHANGED")
			f:RegisterForDrag(true)
			f:SetAttribute("unit", frame.unit)
			return f
		end,
		functions = {
			Load = function(self)

			end,
			Layout = function(self)
				local sets = self.sets
				if sets.enable ~= true then
					self:Hide()
					self.noUpdate = true
					return
				else
					self:Show()
					self.noUpdate = nil
				end
				self:ClearAllPoints()
				self:SetPoint(sets.anchor, self:GetParent(), sets.x, sets.y)
				self:SetSize(sets.width, sets.height)
				self:EnableMouse(not sets.clickThru)
				self:EnableMouseWheel(not sets.clickThru)

				self.zoomlevel = self.sets.zoom
				self.rotation = self.sets.rotation

				local lay = layers[sets.frameLayer]
				self:SetFrameStrata(lay)
				self:SetFrameLevel(sets.frameLevel)
--				self:SetAnimation(91)
			end,
			Update = function(self)
				if self.noUpdate then
					return
				end
				local unit = self:GetParent().unit
				local dead = UnitIsGhost(unit) or UnitIsDead(unit)
				local sets = self.sets
				if (max == 0 or dead) and (self:GetAlpha() ~= .5) then
					self:SetAlpha(.5)
				elseif self:GetAlpha() ~= 1 then
					self:SetAlpha(1)
				end
				self.lastUnit = self.lastUnit or UnitGUID(unit)
				if self.lastUnit ~= UnitGUID(unit) then
					self.lastUnit = UnitGUID(unit)
					self:GetScript("OnEvent")(self)
				end
			end,
		},
		scripts = {
			OnEvent = function(self, ...)
				if self:GetParent().unit and UnitName(self:GetParent().unit) then
					self:SetUnit(self:GetParent().unit)
				end
			end,
			OnShow = function(self)
				self:GetScript("OnEvent")(self)
				self:SetRotation(self.rotation or MODELFRAME_DEFAULT_ROTATION)
				self:SetPortraitZoom(self.zoomlevel or MODELFRAME_MIN_ZOOM)
			end,
			OnMouseWheel = function(self, ...)
				Model_OnMouseWheel(self, ...)
				if self.zoomLevel ~= self.sets.zoom then
					self.sets.zoom = self.zoomLevel
				end
			end,
			OnMouseUp = function(self, ...)
				Model_OnMouseUp(self, ...)
				if self.rotation ~= self.sets.rotation then
					self.sets.rotation = self.rotation
				end
			end,
		},
		options = {
			{
				name = "Anchor",
				kind = "Menu",
				key = "anchor",
				table = anchors,
			},
			{
				name = "Enable",
				kind = "CheckButton",
				key = "enable",
			},
			{
				name = "Death Effect",
				kind = "CheckButton",
				key = "deathEffect",
			},
			{
				name = "Click Thru",
				kind = "CheckButton",
				key = "clickThru",
			},
			{
				name = "X Offset",
				kind = "Slider",
				key = "x",
				min = -400,
				max = 400,
			},
			{
				name = "Y Offset",
				kind = "Slider",
				key = "y",
				min = -400,
				max = 400,
			},
			{
				name = "Width",
				kind = "Slider",
				key = "width",
				min = 10,
				max = 400,
			},
			{
				name = "Height",
				kind = "Slider",
				key = "height",
				min = 10,
				max = 400,
			},
			{
				name = "Scale",
				kind = "Slider",
				key = "scale",
				min = 25,
				max = 250,
			},
			{
				name = "Frame Level",
				kind = "Slider",
				key = "frameLevel",
				min = 1,
				max = 100,
			},	
			{
				name = "Frame Layer",
				kind = "Slider",
				key = "frameLayer",
				min = 1,
				max = 8,
			},
		},
		defaults = {
			zoom = 0.7,
			frameLevel = 1,
			frameLayer = 1,
			scale = 1,
			justifyV = "CENTER",
			width = 133,
			y = -45,
			x = -110,
			justifyH = "CENTER",
			rotation = 0.449085388183594,
			anchor = "BottomLeft",
			height = 197,
			enable = false,
			clickThru = true,
		},
	},--model
	{
		name = "buffs",
		create = function(frame)
			local name = frame:GetName().."Buffs"
			local f = _G[name] or CreateFrame("Frame", name, frame.components.box)
			f.drop = f.drop or CreateFrame("Frame", f:GetName().."backdrop", f)
			f.drop:SetFrameLevel(f:GetFrameLevel()-1)
			f.drop:SetAllPoints(f)
			f.filter = "HELPFUL"
			f.kind = "buff"
			f:SetAttribute("unit", frame.unit)
			f:SetAttribute("filter", f.filter)
			RegisterUnitWatch(self)
			f.unit = frame.unit
			return f
		end,
		functions = {
			Load = function(self)
				self.auras = self.auras or {}
				self.total = self.total or 0
			end,
			Layout = function(self)
				local sets = self.sets
				if sets.enable ~=true then
					self:Hide()
					self.noUpdate = true
					return
				else
					--self:Show()
					self.noUpdate = nil
				end
				self:ClearAllPoints()
				local x = sets.x/(sets.scale/100)
				local y = sets.y/(sets.scale/100)
				self:SetPoint(sets.anchor, self:GetParent(), x, y)
				if sets.scale < 25 then
					sets.scale = sets.scale *100
				end
				self:SetScale(sets.scale/100)
				self:SetAlpha(sets.alpha/100)
					
				for i = 1, sets.rows * sets.columns do
					self:getAura(i)
				end
		
				local w, h = self.auras[1]:GetSize()
				w, h =  w + sets.spacing,  h + sets.spacing
		
				local newWidth =  w * sets.columns - sets.spacing
				local newHeight = h * sets.rows    - sets.spacing
				self:SetSize(newWidth, newHeight)

				local cols, rows = sets.columns, sets.rows
				local isLeftToRight = sets.isLeftToRight
				local isTopToBottom = sets.isTopToBottom
		
				for i, aura in pairs(self.auras) do
					local col, row = (cols-1) - (i-1) % cols, rows - ceil(i / cols)
					if isLeftToRight then
						col = (i-1) % cols
					end
					if isTopToBottom then
						row = ceil(i / cols) - 1
					end
					aura:ClearAllPoints()
					aura:SetPoint('TOPLEFT', w*col, -(h*row))
					aura:Show()
				end
				self:Update()
				if sets.showBG then
					self.drop:SetBackdrop(self.bg)
					self.drop:SetBackdropColor(self.sets.color.r, self.sets.color.g, self.sets.color.b, self.sets.color.a)
				else
					self.drop:SetBackdrop(nil)
				end
			end,
			storeAura = function(self, ...)
				tinsert(self.store, {...})
			end,
			Update = function(self)
				if self.noUpdate then
					return
				end
				local sets = self.sets
				local max = sets.rows * sets.columns
				self.total = 0
				self.store = self.store or {}
				wipe(self.store)
				for i = 1, max do
					local aura = self:getAura(i)
					local name, _, icon, count, debuffType, remaining, expiration = UnitAura(self.unit, i, self.filter);
					if name then
						self.total = self.total + 1
						if count == 0 then
							count = ""
						end
						self:storeAura(icon, count, expiration, remaining)
					else
						break
					end					
				end

				for i = 1, self.total do
					local store = self.store[i]
					if sets.reverse then
						store = self.store[(self.total + 1 ) - i]
					end
					self:updateAura(i, unpack(store))
				end

				for i = self.total+1, #self.auras do
					self:resetAura(i)
				end
				self:updateTooltip()
			end,
			HasMouse = function(self)
				local index
				if self.lastMouse and MouseIsOver(self:getAura(self.lastMouse)) then
					return true, self.lastMouse
				else
					for i = 1, self.total do
						local aura = self:getAura(i)
						if MouseIsOver(aura) and self.isShown then
							self.hasMouse = i
							return true, i
						end
					end
				end
			end,
			updateTooltip = function(self)
				local hasMouse, index =  self:HasMouse()
				if hasMouse then
					if (index ~= self.lastMouse) and self:getAura(index).isShown then
						GameTooltip:SetOwner(self);
						GameTooltip:SetUnitAura(self.unit, self:GetIndex(index), self.filter)
						GameTooltip:ClearAllPoints()
						GameTooltip:SetPoint("BottomRight", self:getAura(index), "TopLeft")
						self.lastMouse = index
					end
				else
					if GameTooltip:GetOwner() == self then
						self.hasMouse = nil
						self.lastMouse = nil
						GameTooltip:Hide()
						return
					end
				end
			end,
			GetIndex = function(self, i)
				local rev = i
				if self.sets.reverse then
					rev = (self.total + 1 ) - i
				end
				return rev
			end,
			updateAura = function(self, i, icon, count, expiration, remaining)
				local aura = self:getAura(i)
				if aura.lastIcon ~=icon then --only make a change, if there is a change.
					aura.lastIcon = icon
					aura:SetTexture(icon)
					aura:SetTexCoord(.1,.9,.1,.9)
				end
		
				if (aura.lastExpiration ~= expiration) then
					aura.lastExpiration = expiration
					CooldownFrame_SetTimer(aura.cooldown, expiration - remaining, remaining, 1)
				end
				if aura.lastCount ~= count then
					aura.lastCount = count
					aura.count:SetText(count)
				end
				if not aura.isShown then
					aura.isShown = true
				end
			end,
			resetAura = function(self, i)
				local aura = self:getAura(i)
				if aura.isShown then
					aura.lastIcon = nil
					aura.lastRemaining = nil
					aura.lastCount = nil
					aura:SetTexture("")
					aura.cooldown:Hide()
					aura.count:SetText("")
					aura.isShown = nil
					if MouseIsOver(aura) then
						self.hasMouse = nil
						self.lastMouse = nil
					end
				end
			end,
			bg = {
				bgFile = 'Interface\\Tooltips\\UI-Tooltip-Background',
				insets = {left = 0, right = 0, top = 0, bottom = 0},
				tile = false,
			},
			getAura = function(self, i)
				local t = self.auras[i] or self:CreateTexture(self:GetName()..self.kind..i, "ARTWORK")
				if not self.auras[i] then
					self.auras[i] = t
					t:SetSize(32, 32)
					t.cooldown = CreateFrame("Cooldown", t:GetName().."Cooldown", self, "CooldownFrameTemplate")
					t.cooldown:SetAllPoints(t)
					t.count = self:CreateFontString(t:GetName().."Count", 'OVERLAY', 'GameFontNormal')
					t.count:SetPoint("BottomRight", t)
					t.isShown = nil
					t.index = i
				end
				return t
			end,
		},
		options = {
			{
				name = "Anchor",
				kind = "Menu",
				key = "anchor",
				table = anchors,
			},
			{
				name = "Enable",
				kind = "CheckButton",
				key = "enable",
			},
			{
				name = "Reverse",
				kind = "CheckButton",
				key = "reverse",
			},
			{
				name = "isTopToBottom",
				kind = "CheckButton",
				key = "isTopToBottom",
			},
			{
				name = "isLeftToRight",
				kind = "CheckButton",
				key = "isLeftToRight",
			},
			{
				name = "Backdrop",
				kind = "CheckButton",
				key = "showBG",
			},
			{
				name = "Backdrop Color",
				kind = "ColorPicker",
				key = "color",
			},
			{
				name = "X Offset",
				kind = "Slider",
				key = "x",
				min = -400,
				max = 400,
			},
			{
				name = "Y Offset",
				kind = "Slider",
				key = "y",
				min = -400,
				max = 400,
			},
			{
				name = "Columns",
				kind = "Slider",
				key = "columns",
				min = 1,
				max = 10,
			},
			{
				name = "Rows",
				kind = "Slider",
				key = "rows",
				min = 1,
				max = 10,
			},
			{
				name = "Spacing",
				kind = "Slider",
				key = "spacing",
				min = -13,
				max = 30,
			},
			{
				name = "Scale",
				kind = "Slider",
				key = "scale",
				min = 25,
				max = 250,
			},
			{
				name = "Opacity",
				kind = "Slider",
				key = "alpha",
				min = 0,
				max = 100,
			},	
		},
		defaults = {
			columns = 6,
			scale = 76,
			rows = 3,
			alpha = 100,
			y = 73,
			x = 0,
			isLeftToRight = true,
			spacing = 0,
			anchor = "Bottom",
			enable = false,
			showBG = false,
			isTopToBottom = false,
			color = {
				a = 1,
				r = 0,
				g = 0,
				b = 0,
			},
		},
	},--buffs
	{
		name = "debuffs",
		create = function(frame)
			local name = frame:GetName().."Debuffs"
			local f = _G[name] or CreateFrame("Frame", name, frame.components.box)
			f.drop = f.drop or CreateFrame("Frame", f:GetName().."backdrop", f)
			f.drop:SetFrameLevel(f:GetFrameLevel()-1)
			f.drop:SetAllPoints(f)
			f.filter = "HARMFUL"
			f.kind = "debuff"
			f:SetAttribute("unit", frame.unit)
			f:SetAttribute("filter", f.filter)
			RegisterUnitWatch(self)
			f.unit = frame.unit
			return f
		end,
		functions = {
			Load = function(self)
				self.auras = self.auras or {}
				self.total = self.total or 0
			end,
			Layout = function(self)
				local sets = self.sets
				if sets.enable ~=true then
					self:Hide()
					self.noUpdate = true
					return
				else
					--self:Show()
					self.noUpdate = nil
				end
				self:ClearAllPoints()
				local x = sets.x/(sets.scale/100)
				local y = sets.y/(sets.scale/100)
				self:SetPoint(sets.anchor, self:GetParent(), x, y)
				if sets.scale < 25 then
					sets.scale = sets.scale *100
				end
				self:SetScale(sets.scale/100)
				self:SetAlpha(sets.alpha/100)
					
				for i = 1, sets.rows * sets.columns do
					self:getAura(i)
				end
		
				local w, h = self.auras[1]:GetSize()
				w, h =  w + sets.spacing,  h + sets.spacing
		
				local newWidth =  w * sets.columns - sets.spacing
				local newHeight = h * sets.rows    - sets.spacing
				self:SetSize(newWidth, newHeight)

				local cols, rows = sets.columns, sets.rows
				local isLeftToRight = sets.isLeftToRight
				local isTopToBottom = sets.isTopToBottom
		
				for i, aura in pairs(self.auras) do
					local col, row = (cols-1) - (i-1) % cols, rows - ceil(i / cols)
					if isLeftToRight then
						col = (i-1) % cols
					end
					if isTopToBottom then
						row = ceil(i / cols) - 1
					end
					aura:ClearAllPoints()
					aura:SetPoint('TOPLEFT', w*col, -(h*row))
					aura:Show()
				end
				self:Update()
				if sets.showBG then
					self.drop:SetBackdrop(self.bg)
					self.drop:SetBackdropColor(self.sets.color.r, self.sets.color.g, self.sets.color.b, self.sets.color.a)
				else
					self.drop:SetBackdrop(nil)
				end
			end,
			storeAura = function(self, ...)
				tinsert(self.store, {...})
			end,
			Update = function(self)
				if self.noUpdate then
					return
				end
				local sets = self.sets
				local max = sets.rows * sets.columns
				self.total = 0
				self.store = self.store or {}
				wipe(self.store)
				for i = 1, max do
					local aura = self:getAura(i)
					local name, _, icon, count, debuffType, remaining, expiration = UnitAura(self.unit, i, self.filter);
					if name then
						self.total = self.total + 1
						if count == 0 then
							count = ""
						end
						self:storeAura(icon, count, expiration, remaining)
					else
						break
					end					
				end

				for i = 1, self.total do
					local store = self.store[i]
					if sets.reverse then
						store = self.store[(self.total + 1 ) - i]
					end
					self:updateAura(i, unpack(store))
				end

				for i = self.total+1, #self.auras do
					self:resetAura(i)
				end

				self:updateTooltip()
			end,
			
			HasMouse = function(self)
				local index
				if self.lastMouse and MouseIsOver(self:getAura(self.lastMouse)) then
					return true, self.lastMouse
				else
					for i = 1, self.total do
						local aura = self:getAura(i)
						if MouseIsOver(aura) then
							self.hasMouse = i
							return true, i
						end
					end
				end
			end,
			updateTooltip = function(self)
				local hasMouse, index =  self:HasMouse()
				if hasMouse then
					if (index ~= self.lastMouse) then
						GameTooltip:SetOwner(self);
						GameTooltip:SetUnitAura(self.unit, self:GetIndex(index), self.filter)
						GameTooltip:ClearAllPoints()
						GameTooltip:SetPoint("BottomRight", self:getAura(index), "TopLeft")
						self.lastMouse = index
					end
				else
					if GameTooltip:GetOwner() == self then
						self.hasMouse = nil
						self.lastMouse = nil
						GameTooltip:Hide()
						return
					end
				end
			end,
			GetIndex = function(self, i)
				local rev = i
				if self.sets.reverse then
					rev = (self.total + 1 ) - i
				end
				return rev
			end,
			
			updateAura = function(self, i, icon, count, expiration, remaining)
				local aura = self:getAura(i)
				if aura.lastIcon ~=icon then --only make a change, if there is a change.
					aura.lastIcon = icon
					aura:SetTexture(icon)
					aura:SetTexCoord(.1,.9,.1,.9)
				end
		
				if (aura.lastExpiration ~= expiration) then
					aura.lastExpiration = expiration
					CooldownFrame_SetTimer(aura.cooldown, expiration - remaining, remaining, 1)
				end
				if aura.lastCount ~= count then
					aura.lastCount = count
					aura.count:SetText(count)
				end
				if not aura.isShown then
					aura.isShown = true
				end
			end,
			resetAura = function(self, i)
				local aura = self:getAura(i)
				if aura.isShown == true then
					aura.lastIcon = nil
					aura.lastRemaining = nil
					aura.lastCount = nil
					aura:SetTexture("")
					aura.cooldown:Hide()
					aura.count:SetText("")
					aura.isShown = false
				end
			end,
			bg = {
				bgFile = 'Interface\\Tooltips\\UI-Tooltip-Background',
				insets = {left = 0, right = 0, top = 0, bottom = 0},
				tile = false,
			},
			getAura = function(self, i)
				local t = self.auras[i] or self:CreateTexture(self:GetName()..self.kind..i, "ARTWORK")
				if not self.auras[i] then
					self.auras[i] = t
					t:SetSize(32, 32)
					t.cooldown = CreateFrame("Cooldown", t:GetName().."Cooldown", self, "CooldownFrameTemplate")
					t.cooldown:SetAllPoints(t)
					t.count = self:CreateFontString(t:GetName().."Count", 'OVERLAY', 'GameFontNormal')
					t.count:SetPoint("BottomRight", t)
					t.isShown = false
					t.index = i
				end
				return t
			end,
		},
		options = {
			{
				name = "Enable",
				kind = "CheckButton",
				key = "enable",
			},
			{
				name = "Anchor",
				kind = "Menu",
				key = "anchor",
				table = anchors,
			},
			{
				name = "Reverse",
				kind = "CheckButton",
				key = "reverse",
			},
			{
				name = "Flip Horizontal",
				kind = "CheckButton",
				key = "isTopToBottom",
			},
			{
				name = "Flip Vertical",
				kind = "CheckButton",
				key = "isLeftToRight",
			},
			{
				name = "Backdrop Color",
				kind = "ColorPicker",
				key = "color",
			},
			{
				name = "Backdrop",
				kind = "CheckButton",
				key = "showBG",
			},


			{
				name = "X Offset",
				kind = "Slider",
				key = "x",
				min = -400,
				max = 400,
			},
			{
				name = "Y Offset",
				kind = "Slider",
				key = "y",
				min = -400,
				max = 400,
			},
			{
				name = "Opacity",
				kind = "Slider",
				key = "alpha",
				min = 0,
				max = 100,
			},
			{
				name = "Spacing",
				kind = "Slider",
				key = "spacing",
				min = -13,
				max = 30,
			},
			{
				name = "Columns",
				kind = "Slider",
				key = "columns",
				min = 1,
				max = 10,
			},
			{
				name = "Rows",
				kind = "Slider",
				key = "rows",
				min = 1,
				max = 10,
			},
			{
				name = "Scale",
				kind = "Slider",
				key = "scale",
				min = 25,
				max = 250,
			},
		},
		defaults = {
			columns = 6,
			scale = 76,
			rows = 3,
			alpha = 100,
			y = -73,
			x = 0,
			isLeftToRight = true,
			spacing = 0,
			anchor = "Top",
			enable = false,
			showBG = false,
			isTopToBottom = true,
			color = {
				a = 1,
				r = 0,
				g = 0,
				b = 0,
			},
		},
	},--debuffs
}

local L

local unitBar = Dominos:CreateClass('Frame', Dominos.Frame)
Dominos.unitBar = unitBar
unitBar.updates = {}

local frames = {}

function unitBar:New(unit)
	local f = self.super.New(self, string.lower(unit))
--	wipe(f.sets)
	f.unit = string.lower(unit)
	f:Reload()
	return f
end

local defaults = {
	focus = {
		point = "Center",
		scale = 1,
		padW = 6,
		width = 140,
		y = 48,
		x = 272,
		anchor = "playerBC",
		padH = 6,
		height = 70,
	},
	targettarget = {
		scale = .75,
		padW = 5,
		anchor = "targetRB",
		width = 140,
		y = 0,
		x = 0,
		padH = 5,
		height = 70,
	},
	target = {
		point = "Center",
		scale = 1,
		padW = 0,
		width = 140,
		y = -152,
		x = 312,
		padH = 0,
		height = 70,
		buffs = {
			enable = true,
		},
		debuffs = {
			enable = true,
		},
	},
	player = {
		point = "Center",
		scale = 1,
		padW = 0,
		width = 140,
		y = -152,
		x = -312,
		padH = 0,
		height = 70,
	},
	focustarget = {
		point = "Center",
		scale = .75,
		padW = 6,
		anchor = "focusBR",
		width = 140,
		y = 134,
		x = 272,
		padH = 6,
		height = 70,
	},
}

function unitBar:GetDefaults()
	local def = {
		point = "Center",
		scale = 1,
		padW = 0,
		width = 140,
		y = 0,
		x = 0,
		padH = 0,
		height = 70,
	}
	if self.unit and defaults[self.unit] then 
		def = CopyTable(defaults[self.unit])
	 end
	return def
end

local function check(source, target)
	--you may now add new defaults at will. ~Goranaws
	for key, value in pairs(source) do
		if type(value) == "table" then
			target[key] = target[key] or check(value, {})
		else
			if (type(value) == "boolean") then
				if target[key] == nil then
					target[key] = value
				end
 			else
 				target[key] = target[key] or value
 			end
		end
	end
	return target
end

function unitBar:Reload() --pretty... I like this idea!			~Goranaws
	check(self:GetDefaults(), self.sets)
	self.components = self.components or {}
	for i, b in pairs(components) do
		self.components[b.name] = self.components[b.name] or self:getComponent(self, b.name, b)
		self.sets[b.name] = self.sets[b.name] or {}
	end
	for key, part in pairs(components) do
		if part.defaults then
			check(part.defaults, self.sets[part.name])
			self.components[part.name].sets = self.sets[part.name]
		end
	end
	self:SetParent(self.parent or self:GetParent())
	self:Layout()
end

function unitBar:Disable()
	self:Free()
	self.parent = self.parent or self:GetParent()
	self:SetParent(hideBlizz)
end

function unitBar:Layout()
	local pad = self:GetPadding()
	local w, h = self.sets.width + pad, self.sets.height + pad
	self:SetSize(w, h)
	for key, part in pairs(self.components) do
		if part.Layout then
			part:Layout()
		end
	end
end

function unitBar:getComponent(parent, i, b)
	local c = self.components[i] or check( b.functions, b.create(parent))
	if b.scripts then
		for i, b in pairs(b.scripts) do
			c:SetScript(i, b)
		end
	end

	if c.Load then
		c:Load()
	end

	self.components[i] = c
	return c
end

local create = {}

function create.Slider(panel, name, key, min, max, step)
	return panel:NewSlider(name, min, max, step or 1,
		function(self)
			self:SetValue(self:GetParent().owner.sets[panel.name][key])
		end,
		function(self) 
			local owner = self:GetParent().owner
			owner.sets[panel.name][key] = self:GetValue()
			owner:Layout()
		end
	)
end

function create.CheckButton(panel, name, key)
	local c = panel:NewCheckButton(name)
	c:SetScript("OnShow",
		function(self)
			self:SetChecked(self:GetParent().owner.sets[panel.name][key])
		end
	)

	c:SetScript("OnClick",
		function(self)
			local owner = self:GetParent().owner
			owner.sets[panel.name][key] = self:GetChecked() or false
			owner:Layout()
		end
	)
	return c
end

local function newButton(menu, name, isColor)
	local b = CreateFrame("Button", menu:GetName()..name, menu, "UIMenuButtonStretchTemplate")
	b:SetSize(140, 25)
	b:SetText(name)
	b.text = b:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	b.text:SetDrawLayer(BACKGROUND)
	if isColor then
		b.texture = b:CreateTexture(nil, "BACKGROUND")
		b.texture:SetDrawLayer("BACKGROUND", -8)
		b.texture:SetTexture('Interface\\Tooltips\\UI-Tooltip-Background')
	else
		b.texture = CreateFrame("Frame", b:GetName().."texture", b)
		b.texture:SetFrameLevel(1)
	end
	
	b.text:SetDrawLayer("ARTWORK", 7)

	
	
	b.texture:SetPoint("TopLeft", b, "BottomLeft")
	b.texture:SetPoint("TopRight", b, "BottomRight")
	b.texture:SetHeight(25)
	b.text:SetAllPoints(b.texture)
	local prev = menu.checkbutton
	if prev then
		b:SetPoint('TopLeft',  prev, 'BottomLeft', 0, -2)
	else
		b:SetPoint('TOPLEFT', 2, 0)
	end
	menu.checkbutton = b
	menu.height = menu.height + 52
	return b
end

function create.EditBox(p, name, key, setter)
	local editBox = CreateFrame('EditBox', p:GetName() .. name, p,  'InputBoxTemplate')
	editBox:SetHeight(20)
	editBox:SetAutoFocus(false)

	local set = CreateFrame('Button', editBox:GetName()..'Set', p, 'UIPanelButtonTemplate')
	
	local flip, flop = name, "Set"
	
	local hold = false
	local isLocked = false
	local function flipText(why)
		local text = flip
		if why == 'lock' then 
			isLocked = true
		elseif why == 'unlock' then 
			isLocked = false
		end
		
		if isLocked == true then
			text = flop
		else
			text = flip
		end
		
		if (why == 'OnEnter') then
			if isLocked == true then
				text = flip
			else
				text = flop
			end
		elseif (why == 'OnLeave')  then
			if isLocked == true then
				text = flop
			else
				text = flip
			end
		end 
		set:SetText(text)
	end

	set:SetScript('OnEnter', function(self)
		flipText('OnEnter')
	end)
	
	set:SetScript('OnLeave', function(self)
		flipText('OnLeave')
	end)

	editBox:SetScript('OnShow', function(self)
		self:SetText(self:GetParent().owner.sets[p.name][key] or '')
	end)
	
	editBox:SetScript('OnEscapePressed', function(self)
		self:SetText(self:GetParent().owner.sets[p.name][key] or '')
		self:ClearFocus()
	end)
	editBox:SetScript('OnEnterPressed', function(self)
		local text = self:GetText()
		setter(self:GetParent().owner, text ~= '' and text or nil)
		self:SetText(self:GetParent().owner.sets[p.name][key] or '')
		self:ClearFocus()
	end)
	editBox:SetScript('OnEditFocusLost', function(self) self:HighlightText(0, 0) flipText('unlock') end)
	editBox:SetScript('OnEditFocusGained', function(self) self:HighlightText() flipText('lock') end)
	set:SetHeight(20)
	set:SetText(name)
	set:SetScript('OnClick', function(self)
		editBox:GetScript('OnEnterPressed')(editBox)
	end)
	
	editBox:SetPoint('Left', set, 'Right', 0, 0)
	editBox:SetPoint("Right", p)

	local prev = p.checkbutton
	if prev then
		set:SetPoint('TOP', prev, 'BOTTOM', 0, -2)
	else
		set:SetPoint('TOPLEFT', 3, 0)
	end
	p.checkbutton = set
	
	
	p.height = p.height + 20
	return p
end

function create.Menu(menu, name, key, table)
	local s
	local f = CreateFrame("Frame", menu:GetName()..name, menu)
	f:SetSize(24, 24)
	f.button = CreateFrame("Button", f:GetName().."Button", f)
	f.button:SetPoint("Top")
	f.button:SetSize(24, 24)
	f.button:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Up")
	f.button:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Down")
	f.button:SetDisabledTexture("Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-Disabled")
	f.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")

	f.text = f:CreateFontString(f:GetName() .. 'Text', "OVERLAY", "GameFontHighlightSmall")
	f.text:SetPoint("BottomLeft", f.button, "BottomRight", 6, 2)
	f.text:SetJustifyH('LEFT')

	local title = f:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
	title:SetPoint("TopLeft", f.button, "TopRight", 4, -2)
	title:SetText(name)	

	f:SetScript('OnShow', function(self)
		if self~= f then
			return
		end
		f:initialize()
		f.text:SetText(f:GetParent().owner.sets[menu.name][key])
	end)

	f:SetScript("OnHide", function() CloseDropDownMenus() end)

	f.button:SetScript("OnClick", function(self)
		ToggleDropDownMenu(1, nil, f, "cursor");
		PlaySound("igMainMenuOptionCheckBoxOn");
	end)

	function f:initialize()
		local owner = f:GetParent().owner
		local info = UIDropDownMenu_CreateInfo()
		for i, anchor in ipairs(table) do
			wipe(info)
			info.text = anchor
			info.func = function(item, name)
				owner.sets[menu.name][key] = name
				owner:Layout()
				f.text:SetText(name)
			end
			info.checked = (anchor == owner.sets[menu.name][key])
			info.arg1 = anchor
			UIDropDownMenu_AddButton(info)
		end
	end
	
	local dropDownList = _G["DropDownList"..1]
	dropDownList.dropdown = f;
	dropDownList.shouldRefresh = true;
	
	local prev = menu.checkbutton
	if prev then
		f:SetPoint('TOP', prev, 'BOTTOM', 0, -0)
	else
		f:SetPoint('TOPLEFT', 2, -5)
	end
	f.point = {f:GetPoint()}
	menu.checkbutton = f.button

	menu.height = menu.height + 24
	return f

end

local function ShowColorPicker(r, g, b, a, changedCallback)
	ColorPickerFrame.hasOpacity, ColorPickerFrame.opacity = nil, nil;
	ColorPickerFrame.func, ColorPickerFrame.opacityFunc, ColorPickerFrame.cancelFunc = nil, nil, nil;
	ColorPickerFrame:SetColorRGB(r,g,b);
	ColorPickerFrame.hasOpacity, ColorPickerFrame.opacity = (a ~= nil), a;
	ColorPickerFrame.previousValues = {r,g,b,a};
	ColorPickerFrame.func, ColorPickerFrame.opacityFunc, ColorPickerFrame.cancelFunc = changedCallback, changedCallback, changedCallback;
	ColorPickerFrame:Hide(); -- Need to run the OnShow handler.
	ColorPickerFrame:Show();
end

function create.ColorPicker(menu, name, key)
	local button = CreateFrame("Button", menu:GetName()..name, menu)
	button:SetSize(24, 24)

	button:SetPushedTexture("Interface\\Buttons\\UI-Quickslot-Depress")
	button:SetHighlightTexture("Interface\\Buttons\\ButtonHilight-Square")
	button:GetHighlightTexture():SetBlendMode("ADD")
	button:SetNormalTexture('Interface\\Tooltips\\UI-Tooltip-Background')

	button:SetScript("OnShow", function(self)
		local saved = self:GetParent().owner.sets[string.lower(menu.name)][key]

		button:GetNormalTexture():SetVertexColor(saved.r, saved.g, saved.b, saved.a)
	end)

	button:SetScript("OnClick", function(self)
		local saved = self:GetParent().owner.sets[string.lower(menu.name)][key]
		local r, g, b, a = saved.r, saved.g, saved.b, saved.a
		ShowColorPicker(r, g, b, a, function(restore)
			local newR, newG, newB, newA;
			if restore then
				newR, newG, newB, newA = unpack(restore);
			else
				newA, newR, newG, newB = OpacitySliderFrame:GetValue(), ColorPickerFrame:GetColorRGB();
			end
			saved.r, saved.g, saved.b, saved.a = newR, newG, newB, newA
			self:GetScript("OnShow")(self)
			menu.owner:Layout()
		end)
	end)

--[[
	local b = newButton(menu, name, 1)

	b:SetScript("OnShow", function(self)
		local saved = self:GetParent().owner.sets[string.lower(menu.name)][key]
		b.texture:SetVertexColor(saved.r, saved.g, saved.b, saved.a)
	end)
	b:SetScript("OnClick", function(self)
		local saved = self:GetParent().owner.sets[string.lower(menu.name)][key]
		local r, g, b, a = saved.r, saved.g, saved.b, saved.a
		ShowColorPicker(r, g, b, a, function(restore)
			local newR, newG, newB, newA;
			if restore then
				newR, newG, newB, newA = unpack(restore);
			else
				newA, newR, newG, newB = OpacitySliderFrame:GetValue(), ColorPickerFrame:GetColorRGB();
			end
			saved.r, saved.g, saved.b, saved.a = newR, newG, newB, newA
			self:GetScript("OnShow")(self)
			menu.owner:Layout()
		end)
	end)
--]]
	local prev = menu.checkbutton
	if prev then
		button:SetPoint('TopLeft',  prev, 'BottomLeft', 0, -2)
	else
		button:SetPoint('TOPLEFT', 2, 0)
	end
	menu.checkbutton = button
	menu.height = menu.height + 24
	return button
end

local function AddLayoutPanel(menu)
	local panel = menu:NewPanel(L.Layout)
	panel.opacitySlider = panel:NewOpacitySlider()
	panel.fadeSlider = panel:NewFadeSlider()
	panel.paddingSlider = panel:NewPaddingSlider()

	panel:NewSlider("Height", 10, 600, 1,
		function(self) --Getter
			self:SetValue(self:GetParent().owner.sets.height)
		end,
		function(self) --Setter
			self:GetParent().owner.sets.height = self:GetValue()
			self:GetParent().owner:Layout()
		end
	)

	panel:NewSlider("Width", 10, 600, 1,
		function(self) --Getter
			self:SetValue(self:GetParent().owner.sets.width)
		end,
		function(self) --Setter
			self:GetParent().owner.sets.width = self:GetValue()
			self:GetParent().owner:Layout()
		end
	)
	
	panel.scaleSlider = panel:NewScaleSlider()

	return panel
end

local function AddBasicPanel(menu, name, options)
	local key = string.lower(name)
	local panel = menu:NewPanel(name)
	for name, info in pairs(options) do
		local func = create[info.kind]
		o = func(panel, info.name, info.key, info.min or info.table, info.max, info.step)
	end
	return panel
end

function unitBar:CreateMenu()
	local menu = Dominos:NewMenu(self.id)
	L = LibStub('AceLocale-3.0'):GetLocale('Dominos-Config')
	local panel = AddLayoutPanel(menu)
	for name, info in pairs(components) do
		if info.options then
			AddBasicPanel(menu, info.name, info.options)
		end
	end
	menu:AddAdvancedPanel()
	self.menu = menu
end


--viable unit check
local base = {
	"player",
	"target",
	"pet",
	"focus",
	"arena%d+",
	"boss%d+",
	"party%d+",
	"partypet%d+",
	"raid%d+",
	"raidpet%d+",
	"mouseover",
}

local max = {
	raid = 40,
	raidpet = 40,
	arena = 4,
	boss = 4,
	party = 4,
	partypet = 4,
}

local function IsViable(unit)
	local keep = unit
	unit = string.lower(unit)
	local count = 0
	if string.match (unit, "target" , 2) then
		_, count = string.gsub(string.sub (unit, 2, string.len(unit)), "target", "")
		if (unit == "player") and count > 0 then
			return nil
		else
			unit = string.gsub (unit, "target", "", count)
		end
	end	
	local can
	for i, test in pairs(base) do
		if string.match(unit, test) then
			return string.lower(keep)	
		end
	end
end

--[[ Unit Bar Controller ]]

local unitBarController = Dominos:NewModule('unitBar')

local units = {
	player = true,
	target = true,
	targetTarget = true,
	focus = true,
	focustarget = true,
}

local sets

local function getSets()
	sets = sets or Dominos_Units[UnitName("player")] or {frames = units, disable = {PlayerFrame = true, TargetFrame = true, FocusFrame = true}}
	return sets
end

function unitBarController:Load()
	Dominos_Units = Dominos_Units or {}
	Dominos_Units[UnitName("player")] = getSets()
	
	if #frames > 0 then
		for i, frame in pairs(frames) do
			frame:Reload()
		end
		return
	end

	for unit, b in pairs(getSets().frames) do
		local frame  = frames[string.lower(unit)] or unitBar:New(string.lower(unit))
		frames[string.lower(unit)] = frame
	end
	
	for frameName, b in pairs(getSets().disable) do
		if frameName and _G[frameName] then
			_G[frameName]:SetParent(hideBlizz)
		end
	end
end

function unitBarController:Unload()
	if frames then
		for i, frame in pairs(frames) do
			if not getSets().frames[frame.unit] then
				frame:Disable()
			end
		end
	end
end

local function add(unit)
	if not IsViable(unit) then
		print("invalid unit! :(", unit)
		return
	end
	getSets()[unit] = true
	ReloadUI()
end

local function remove(unit)
	if not IsViable(unit) then
		print("invalid unit! :(", unit)
		return
	end
	getSets()[unit] = nil
	ReloadUI()
end

local function disable(frameName)
	if _G[frameName] then
		getSets().disable[frameName] = true
		_G[frameName]:SetParent(hideBlizz)
	end
end

local function enable(frameName)
	if _G[frameName] then
		getSets().disable[frameName] = nil
		_G[frameName]:SetParent(UIParent)
	end
end

SLASH_DOMUNITS1 = '/dunit'
local function handler(msg)
	local args = {}
	local f
	while string.find(msg, ' ') do
		f = string.split(' ', msg) or msg
		msg = string.gsub(msg, f..' ', "")
		tinsert(args, f)
	end
	local arg1  = unpack(args)
	arg1 = string.lower(arg1)
	if (arg1 == "add") then
		add( string.lower(msg))
	elseif  (arg1 == "remove") then
		remove( string.lower(msg))
	elseif  (arg1 == "enable") then
		enable(msg)
	elseif  (arg1 == "disable") then
		disable(msg)
	end
end
SlashCmdList['DOMUNITS'] = handler
